r"""

Collection of functions for basic visualization of images
#########################################################

This module contains functions for quickly visualizing 3D images in 2D
views.

.. currentmodule:: porespy

.. autosummary::
   :template: mybase.rst
   :toctree: generated/

    visualization.bar
    visualization.imshow
    visualization.prep_for_imshow
    visualization.satn_to_movie
    visualization.satn_to_panels
    visualization.sem
    visualization.set_mpl_style
    visualization.show_3D
    visualization.show_mesh
    visualization.show_planes
    visualization.xray

"""

from ._views import *
from ._plots import *
from ._funcs import *
