r"""

imagej
######

**Access ImageJ Filters**

.. currentmodule:: porespy

.. autosummary::
   :template: mybase.rst
   :toctree: generated/

    filters.imagej.imagej_wrapper
    filters.imagej.imagej_plugin

"""

from ._funcs import imagej_wrapper
from ._funcs import imagej_plugin
